# -*- coding: utf-8 -*-

import logging
from typing import Optional

from evernote_backup.cli_app_auth import get_auth_token, get_sync_client
from evernote_backup.cli_app_storage import (
    get_storage,
    initialize_storage,
    raise_on_existing_database,
    raise_on_old_database_version,
)
from evernote_backup.cli_app_util import ProgramTerminatedError
from evernote_backup.config import CURRENT_DB_VERSION
from evernote_backup.note_exporter import NoteExporter, NothingToExportError
from evernote_backup.note_synchronizer import NoteSynchronizer, WrongAuthUserError

logger = logging.getLogger(__name__)


def init_db(
    database: str,
    auth_user: Optional[str],
    auth_password: Optional[str],
    auth_is_oauth: bool,
    auth_oauth_port: int,
    auth_oauth_host: str,
    auth_token: Optional[str],
    force: bool,
    backend: str,
    network_retry_count: int,
) -> None:
    if not force:
        raise_on_existing_database(database)

    if not auth_token:
        auth_token = get_auth_token(
            auth_user,
            auth_password,
            auth_is_oauth,
            auth_oauth_port,
            auth_oauth_host,
            backend,
            network_retry_count,
        )

    note_client = get_sync_client(auth_token, backend, network_retry_count, 1)

    storage = initialize_storage(database, force)

    new_user = note_client.user

    storage.config.set_config_value("DB_VERSION", str(CURRENT_DB_VERSION))
    storage.config.set_config_value("USN", "0")
    storage.config.set_config_value("auth_token", auth_token)
    storage.config.set_config_value("user", new_user)
    storage.config.set_config_value("backend", backend)

    logger.info(f"Successfully initialized database for {new_user}!")


def reauth(
    database: str,
    auth_user: Optional[str],
    auth_password: Optional[str],
    auth_is_oauth: bool,
    auth_oauth_port: int,
    auth_oauth_host: str,
    auth_token: Optional[str],
    network_retry_count: int,
) -> None:
    storage = get_storage(database)

    raise_on_old_database_version(storage)

    backend = storage.config.get_config_value("backend")

    if not auth_token:
        auth_token = get_auth_token(
            auth_user,
            auth_password,
            auth_is_oauth,
            auth_oauth_port,
            auth_oauth_host,
            backend,
            network_retry_count,
        )

    note_client = get_sync_client(auth_token, backend, network_retry_count, 1)

    local_user = storage.config.get_config_value("user")

    if local_user != note_client.user:
        raise ProgramTerminatedError(
            f"Current user of this database is {local_user}, not {note_client.user}!"
            " Each user must use a different database file."
        )

    storage.config.set_config_value("auth_token", auth_token)

    logger.info(f"Successfully refreshed auth token for {local_user}!")


def sync(
    database: str,
    max_chunk_results: int,
    max_download_workers: int,
    download_cache_memory_limit: int,
    network_retry_count: int,
) -> None:
    storage = get_storage(database)

    raise_on_old_database_version(storage)

    backend = storage.config.get_config_value("backend")
    auth_token = storage.config.get_config_value("auth_token")

    note_client = get_sync_client(
        auth_token, backend, network_retry_count, max_chunk_results
    )

    note_synchronizer = NoteSynchronizer(
        note_client, storage, max_download_workers, download_cache_memory_limit
    )

    try:
        note_synchronizer.sync()
    except WrongAuthUserError as e:
        raise ProgramTerminatedError(
            f"Current user of this database is {e.local_user}, not {e.remote_user}!"
            " Each user must use a different database file."
        )

    logger.info("Synchronization completed!")


def export(
    database: str,
    single_notes: bool,
    include_trash: bool,
    no_export_date: bool,
    output_path: str,
) -> None:
    storage = get_storage(database)

    raise_on_old_database_version(storage)

    exporter = NoteExporter(storage, output_path)

    try:
        exporter.export_notebooks(single_notes, include_trash, no_export_date)
    except NothingToExportError:
        raise ProgramTerminatedError(
            "Database is empty, nothing to export."
            " Execute sync command to populate database first!"
        )

    logger.info("All notes have been exported!")



NETWORK_ERROR_RETRY_COUNT = 50
OAUTH_LOCAL_PORT = 10500
OAUTH_HOST = "localhost"
SYNC_CHUNK_MAX_RESULTS = 200
SYNC_MAX_DOWNLOAD_WORKERS = 5
SYNC_DOWNLOAD_CACHE_MEMORY_LIMIT = 256
DATABASE_NAME = "en_backup.db"
BACKEND = "evernote"

SYNC_CHUNK_MAX_RESULTS_SERVER_LIMIT = 256
SYNC_MAX_DOWNLOAD_WORKERS_SANE_LIMIT = 256



import base64
import sys
from datetime import datetime
from typing import Optional


def fmt_time(timestamp: Optional[int]) -> Optional[str]:
    if timestamp is None:
        return timestamp

    timestamp //= 1000

    if timestamp < _get_max_timestamp():
        date = datetime.utcfromtimestamp(timestamp)
    else:
        date = _date_from_future(timestamp)

    return date.strftime("%Y%m%dT%H%M%SZ")


def fmt_binary(binary_data: bytes) -> str:
    slice_width = 120
    return (
        "\n"
        + _slice_str(base64.b64encode(binary_data).decode(), slice_width)
        + "\n      "
    )


def fmt_content(content_body: Optional[str]) -> Optional[str]:
    if content_body is None:
        return content_body

    body = content_body.strip()

    # <?xml version="1.0" encoding="UTF-8"?>
    if body.startswith("<?xml") and body.find(">") != -1:
        content_start = body.find(">") + 1
        body = body[content_start:].strip()

    return (
        f"\n      "
        f"<![CDATA["
        f'<?xml version="1.0" encoding="UTF-8" standalone="no"?>\n'
        f"{body}"
        f"]]>\n    "
    )


def _slice_str(text: str, width: int) -> str:
    slice_steps = range(0, len(text), width)
    slice_lines = (text[i : i + width] for i in slice_steps)
    return "\n".join(slice_lines)


# https://github.com/arrow-py/arrow/blob/10e8970baabeefd8c2f31814d66b11421ac128f1/arrow/constants.py
def _get_max_timestamp() -> int:  # pragma: no cover
    try:
        return int(datetime.max.timestamp())
    except (OverflowError, ValueError, OSError):
        is_64bits = sys.maxsize > 2**32  # noqa: WPS114
        return int(
            datetime(3000, 1, 1, 23, 59, 59, 999999).timestamp()
            if is_64bits
            else datetime(2038, 1, 1, 23, 59, 59, 999999).timestamp()
        )


# https://stackoverflow.com/a/42936293/13100286
# https://howardhinnant.github.io/date_algorithms.html#civil_from_days
def _date_from_future(timestamp: int) -> datetime:  # noqa: WPS210
    z = timestamp // 86400 + 719468
    era = (z if z >= 0 else z - 146096) // 146097
    doe = z - era * 146097
    yoe = (doe - doe // 1460 + doe // 36524 - doe // 146096) // 365
    y = yoe + era * 400
    doy = doe - (365 * yoe + yoe // 4 - yoe // 100)
    mp = (5 * doy + 2) // 153
    d = doy - (153 * mp + 2) // 5 + 1
    m = mp + (3 if mp < 10 else -9)
    y += m <= 2

    day_time = datetime.utcfromtimestamp(timestamp % 86400)

    return datetime(
        year=y,
        month=m,
        day=d,
        hour=day_time.hour,
        minute=day_time.minute,
        second=day_time.second,
    )



from datetime import datetime


def get_token_expiration_date(token: str) -> datetime:
    hex_base = 16
    token_parts = token.split(":")
    token_expire_ts = int(token_parts[2][2:], base=hex_base) // 1000

    return datetime.utcfromtimestamp(token_expire_ts)


def get_token_shard(token: str) -> str:
    if not token:
        return ""

    return token[2 : token.index(":")]



import os

from evernote_backup.cli_app_util import ProgramTerminatedError, logger
from evernote_backup.note_storage import (
    DatabaseResyncRequiredError,
    SqliteStorage,
    initialize_db,
)


def get_storage(database_path: str) -> SqliteStorage:
    logger.info("Reading database {0}...".format(os.path.basename(database_path)))

    try:
        return SqliteStorage(database_path)
    except FileNotFoundError:
        raise ProgramTerminatedError(
            f"Database file {database_path} does not exist."
            f" Initialize database first!"
        )


def raise_on_old_database_version(storage: SqliteStorage) -> None:
    try:
        storage.check_version()
    except DatabaseResyncRequiredError:
        raise ProgramTerminatedError(
            "The database version has been updated. Full resync is required."
        )


def raise_on_existing_database(database_path: str) -> None:
    if os.path.exists(database_path):
        raise ProgramTerminatedError(
            "Database already exists."
            " Use --force option to overwrite existing database file."
        )


def initialize_storage(database_path: str, force: bool) -> SqliteStorage:
    logger.info("Initializing database {0}...".format(os.path.basename(database_path)))

    if force:
        if os.path.exists(database_path):
            os.remove(database_path)
    else:
        raise_on_existing_database(database_path)

    initialize_db(database_path)

    return get_storage(database_path)



from evernote.edam.type.ttypes import TType
from evernote.edam.userstore import UserStore
from thrift.protocol import TBinaryProtocol
from thrift.Thrift import TMessageType
from thrift.transport import TTransport

try:
    from thrift.protocol import fastbinary
except ImportError:
    fastbinary = None


class authenticateLongSessionV2_args(object):
    """
    Attributes:
     - usernameOrEmail
     - password
     - ssoLoginToken
     - consumerKey
     - consumerSecret
     - deviceIdentifier
     - deviceDescription
     - supportsTwoFactor
     - supportsBusinessOnlyAccounts
    """

    thrift_spec = (
        None,
        (1, TType.STRING, "usernameOrEmail", None, None),
        (2, TType.STRING, "password", None, None),
        (3, TType.STRING, "ssoLoginToken", None, None),
        (4, TType.STRING, "consumerKey", None, None),
        (5, TType.STRING, "consumerSecret", None, None),
        (6, TType.STRING, "deviceIdentifier", None, None),
        (7, TType.STRING, "deviceDescription", None, None),
        (8, TType.BOOL, "supportsTwoFactor", None, None),
        (9, TType.BOOL, "supportsBusinessOnlyAccounts", None, None),
    )

    def __init__(
        self,
        usernameOrEmail=None,
        password=None,
        ssoLoginToken=None,
        consumerKey=None,
        consumerSecret=None,
        deviceIdentifier=None,
        deviceDescription=None,
        supportsTwoFactor=None,
        supportsBusinessOnlyAccounts=None,
    ):
        self.usernameOrEmail = usernameOrEmail
        self.password = password
        self.ssoLoginToken = ssoLoginToken
        self.consumerKey = consumerKey
        self.consumerSecret = consumerSecret
        self.deviceIdentifier = deviceIdentifier
        self.deviceDescription = deviceDescription
        self.supportsTwoFactor = supportsTwoFactor
        self.supportsBusinessOnlyAccounts = supportsBusinessOnlyAccounts

    def read(self, iprot):
        if (
            iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated
            and isinstance(iprot.trans, TTransport.CReadableTransport)
            and self.thrift_spec is not None
            and fastbinary is not None
        ):  # pragma: no cover
            fastbinary.decode_binary(
                self, iprot.trans, (self.__class__, self.thrift_spec)
            )
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.usernameOrEmail = iprot.readString().decode("utf-8")
                else:  # pragma: no cover
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.password = iprot.readString().decode("utf-8")
                else:  # pragma: no cover
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.STRING:
                    self.ssoLoginToken = iprot.readString().decode("utf-8")
                else:  # pragma: no cover
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.STRING:
                    self.consumerKey = iprot.readString().decode("utf-8")
                else:  # pragma: no cover
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.STRING:
                    self.consumerSecret = iprot.readString().decode("utf-8")
                else:  # pragma: no cover
                    iprot.skip(ftype)
            elif fid == 6:
                if ftype == TType.STRING:
                    self.deviceIdentifier = iprot.readString().decode("utf-8")
                else:  # pragma: no cover
                    iprot.skip(ftype)
            elif fid == 7:
                if ftype == TType.STRING:
                    self.deviceDescription = iprot.readString().decode("utf-8")
                else:  # pragma: no cover
                    iprot.skip(ftype)
            elif fid == 8:
                if ftype == TType.BOOL:
                    self.supportsTwoFactor = iprot.readBool()
                else:  # pragma: no cover
                    iprot.skip(ftype)
            elif fid == 9:
                if ftype == TType.BOOL:
                    self.supportsBusinessOnlyAccounts = iprot.readBool()
                else:  # pragma: no cover
                    iprot.skip(ftype)
            else:  # pragma: no cover
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if (
            oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated
            and self.thrift_spec is not None
            and fastbinary is not None
        ):  # pragma: no cover
            oprot.trans.write(
                fastbinary.encode_binary(self, (self.__class__, self.thrift_spec))
            )
            return
        oprot.writeStructBegin("authenticateLongSession_args")
        if self.usernameOrEmail is not None:
            oprot.writeFieldBegin("username", TType.STRING, 1)
            oprot.writeString(self.usernameOrEmail.encode("utf-8"))
            oprot.writeFieldEnd()
        if self.password is not None:
            oprot.writeFieldBegin("password", TType.STRING, 2)
            oprot.writeString(self.password.encode("utf-8"))
            oprot.writeFieldEnd()
        if self.ssoLoginToken is not None:
            oprot.writeFieldBegin("ssoLoginToken", TType.STRING, 3)
            oprot.writeString(self.ssoLoginToken.encode("utf-8"))
            oprot.writeFieldEnd()
        if self.consumerKey is not None:
            oprot.writeFieldBegin("consumerKey", TType.STRING, 4)
            oprot.writeString(self.consumerKey.encode("utf-8"))
            oprot.writeFieldEnd()
        if self.consumerSecret is not None:
            oprot.writeFieldBegin("consumerSecret", TType.STRING, 5)
            oprot.writeString(self.consumerSecret.encode("utf-8"))
            oprot.writeFieldEnd()
        if self.deviceIdentifier is not None:
            oprot.writeFieldBegin("deviceIdentifier", TType.STRING, 6)
            oprot.writeString(self.deviceIdentifier.encode("utf-8"))
            oprot.writeFieldEnd()
        if self.deviceDescription is not None:
            oprot.writeFieldBegin("deviceDescription", TType.STRING, 7)
            oprot.writeString(self.deviceDescription.encode("utf-8"))
            oprot.writeFieldEnd()
        if self.supportsTwoFactor is not None:
            oprot.writeFieldBegin("supportsTwoFactor", TType.BOOL, 8)
            oprot.writeBool(self.supportsTwoFactor)
            oprot.writeFieldEnd()
        if self.supportsBusinessOnlyAccounts is not None:
            oprot.writeFieldBegin("supportsBusinessOnlyAccounts", TType.BOOL, 9)
            oprot.writeBool(self.supportsBusinessOnlyAccounts)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        """not implemented"""

    def __repr__(self):  # pragma: no cover
        L = ["%s=%r" % (key, value) for key, value in self.__dict__.items()]
        return "%s(%s)" % (self.__class__.__name__, ", ".join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__


class authenticateLongSessionV2_request(object):
    """
    Attributes:
     - auth_args
    """

    thrift_spec = (
        (
            1,
            TType.STRUCT,
            "auth_args",
            (
                authenticateLongSessionV2_args,
                authenticateLongSessionV2_args.thrift_spec,
            ),
            None,
        ),
    )

    def __init__(
        self,
        auth_args=None,
    ):
        self.auth_args = auth_args

    def read(self, iprot):
        if (
            iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated
            and isinstance(iprot.trans, TTransport.CReadableTransport)
            and self.thrift_spec is not None
            and fastbinary is not None
        ):  # pragma: no cover
            fastbinary.decode_binary(
                self, iprot.trans, (self.__class__, self.thrift_spec)
            )
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.auth_args = authenticateLongSessionV2_args()
                    self.auth_args.read(iprot)
                else:  # pragma: no cover
                    iprot.skip(ftype)
            else:  # pragma: no cover
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if (
            oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated
            and self.thrift_spec is not None
            and fastbinary is not None
        ):  # pragma: no cover
            oprot.trans.write(
                fastbinary.encode_binary(self, (self.__class__, self.thrift_spec))
            )
            return
        oprot.writeStructBegin("authenticateLongSessionV2_args")
        if self.auth_args is not None:
            oprot.writeFieldBegin("auth_args", TType.STRUCT, 1)
            self.auth_args.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        """not implemented"""

    def __repr__(self):  # pragma: no cover
        L = ["%s=%r" % (key, value) for key, value in self.__dict__.items()]
        return "%s(%s)" % (self.__class__.__name__, ", ".join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__


class ClientV2(UserStore.Client):
    def authenticateLongSessionV2(
        self,
        username,
        password,
        ssoLoginToken,
        consumerKey,
        consumerSecret,
        deviceIdentifier,
        deviceDescription,
        supportsTwoFactor,
        supportsBusinessOnlyAccounts,
    ):
        self.send_authenticateLongSessionV2(
            username,
            password,
            ssoLoginToken,
            consumerKey,
            consumerSecret,
            deviceIdentifier,
            deviceDescription,
            supportsTwoFactor,
            supportsBusinessOnlyAccounts,
        )
        return self.recv_authenticateLongSession()

    def send_authenticateLongSessionV2(
        self,
        username,
        password,
        ssoLoginToken,
        consumerKey,
        consumerSecret,
        deviceIdentifier,
        deviceDescription,
        supportsTwoFactor,
        supportsBusinessOnlyAccounts,
    ):
        self._oprot.writeMessageBegin(
            "authenticateLongSessionV2", TMessageType.CALL, self._seqid
        )

        args = authenticateLongSessionV2_args()
        args.usernameOrEmail = username
        args.password = password
        args.ssoLoginToken = ssoLoginToken
        args.consumerKey = consumerKey
        args.consumerSecret = consumerSecret
        args.deviceIdentifier = deviceIdentifier
        args.deviceDescription = deviceDescription
        args.supportsTwoFactor = supportsTwoFactor
        args.supportsBusinessOnlyAccounts = supportsBusinessOnlyAccounts

        request = authenticateLongSessionV2_request()
        request.auth_args = args

        request.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()



# -*- coding: utf-8 -*-

import logging
import lzma
import os
import pickle
import sqlite3
from typing import Iterable, Iterator, List, NamedTuple, Optional, Tuple, Union

from evernote.edam.type.ttypes import LinkedNotebook, Note, Notebook

from evernote_backup.config import CURRENT_DB_VERSION
from evernote_backup.log_util import log_format_note, log_format_notebook

logger = logging.getLogger(__name__)


class NoteForSync(NamedTuple):
    guid: str
    title: str
    linked_notebook_guid: Optional[str]


DB_SCHEMA = """CREATE TABLE IF NOT EXISTS notebooks(
                        guid TEXT PRIMARY KEY,
                        name TEXT,
                        stack TEXT
                    );
                    CREATE TABLE IF NOT EXISTS notebooks_linked(
                        guid TEXT PRIMARY KEY,
                        notebook_guid TEXT,
                        usn INT DEFAULT 0
                    );
                    CREATE TABLE IF NOT EXISTS notes(
                        guid TEXT PRIMARY KEY,
                        title TEXT,
                        notebook_guid TEXT,
                        is_active BOOLEAN,
                        raw_note BLOB
                    );
                    CREATE TABLE IF NOT EXISTS config(
                        name TEXT PRIMARY KEY,
                        value TEXT
                    );
                    CREATE INDEX IF NOT EXISTS idx_notes
                     ON notes(notebook_guid, is_active);
                    CREATE INDEX IF NOT EXISTS idx_notes_title
                     ON notes(title COLLATE NOCASE);
                    CREATE INDEX IF NOT EXISTS idx_notebooks_linked
                     ON notebooks_linked(guid, notebook_guid);
                    CREATE INDEX IF NOT EXISTS idx_notes_raw_null
                     ON notes(guid) WHERE raw_note IS NULL;
"""


class DatabaseResyncRequiredError(Exception):
    """Raise when database update requires resync"""


def initialize_db(filename: str) -> None:
    if os.path.exists(filename):
        raise FileExistsError

    db = sqlite3.connect(filename)

    with db as con:
        con.executescript(DB_SCHEMA)

    db.close()


class SqliteStorage(object):
    def __init__(self, database: Union[str, sqlite3.Connection]) -> None:
        if isinstance(database, sqlite3.Connection):
            self.db = database
        else:
            if not os.path.exists(database):
                raise FileNotFoundError("Database file does not exist.")

            self.db = sqlite3.connect(database)
            self.db.row_factory = sqlite3.Row

    @property
    def config(self) -> "ConfigStorage":
        return ConfigStorage(self.db)

    @property
    def notes(self) -> "NoteStorage":
        return NoteStorage(self.db)

    @property
    def notebooks(self) -> "NoteBookStorage":
        return NoteBookStorage(self.db)

    def check_version(self) -> None:
        try:
            db_version = int(self.config.get_config_value("DB_VERSION"))
        except KeyError:
            db_version = 0

        if db_version != CURRENT_DB_VERSION:
            self.upgrade_db(db_version)

    def upgrade_db(self, db_version: int) -> None:
        need_resync = False

        if db_version == 0:
            need_resync = True
            with self.db as con1:
                con1.execute("DROP TABLE notebooks;")
                con1.execute("DROP TABLE notes;")

                con1.executescript(DB_SCHEMA)

        if db_version < 3:
            with self.db as con2:
                con2.execute(
                    "CREATE INDEX IF NOT EXISTS idx_notes_title"
                    " ON notes(title COLLATE NOCASE);"
                )

        if db_version < 4:
            with self.db as con3:
                con3.execute(
                    "CREATE TABLE IF NOT EXISTS notebooks_linked("
                    " guid TEXT PRIMARY KEY,"
                    " notebook_guid TEXT,"
                    " usn INT DEFAULT 0"
                    " );"
                )
                con3.execute(
                    "CREATE INDEX IF NOT EXISTS idx_notebooks_linked"
                    " ON notebooks_linked(guid, notebook_guid);"
                )

        if db_version < 5:
            with self.db as con4:
                con4.execute(
                    "CREATE INDEX IF NOT EXISTS idx_notes_raw_null"
                    " ON notes(guid) WHERE raw_note IS NULL;"
                )

        self.config.set_config_value("DB_VERSION", str(CURRENT_DB_VERSION))

        if need_resync:
            self.config.set_config_value("USN", "0")
            raise DatabaseResyncRequiredError


class NoteBookStorage(SqliteStorage):  # noqa: WPS214
    def add_notebooks(self, notebooks: Iterable[Notebook]) -> None:
        if logger.getEffectiveLevel() == logging.DEBUG:  # pragma: no cover
            for nb in notebooks:
                nb_info = log_format_notebook(nb)
                logger.debug(f"Adding/updating notebook {nb_info}")

        with self.db as con:
            con.executemany(
                "replace into notebooks(guid, name, stack) values (?, ?, ?)",
                ((nb.guid, nb.name, nb.stack) for nb in notebooks),  # noqa: WPS441
            )

    def iter_notebooks(self) -> Iterator[Notebook]:
        with self.db as con:
            cur = con.execute(
                "select guid, name, stack from notebooks",
            )

            yield from (
                Notebook(
                    guid=row["guid"],
                    name=row["name"],
                    stack=row["stack"],
                )
                for row in cur
            )

    def get_notebook_notes_count(self, notebook_guid: str) -> int:
        with self.db as con:
            cur = con.execute(
                "select COUNT(guid) from notes"
                " where notebook_guid=? and is_active=1",
                (notebook_guid,),
            )

            return int(cur.fetchone()[0])

    def expunge_notebooks(self, guids: Iterable[str]) -> None:
        with self.db as con:
            con.executemany("delete from notebooks where guid=?", ((g,) for g in guids))

    def add_linked_notebook(
        self, l_notebook: LinkedNotebook, notebook: Notebook
    ) -> None:
        if logger.getEffectiveLevel() == logging.DEBUG:  # pragma: no cover
            logger.debug(
                f"Adding/updating linked notebook '{l_notebook.shareName}'"
                f" [{l_notebook.guid}] -> [{notebook.guid}]"
            )

        with self.db as con:
            con.execute(
                "replace into notebooks_linked(guid, notebook_guid) values (?, ?)",
                (l_notebook.guid, notebook.guid),
            )

    def get_notebook_by_linked_guid(self, l_notebook_guid: str) -> Notebook:
        with self.db as con:
            cur = con.execute(
                "select notebooks.guid, notebooks.name, notebooks.stack"
                " from notebooks_linked"
                " join notebooks"
                " on notebooks.guid=notebooks_linked.notebook_guid"
                " where notebooks_linked.guid=?",
                (l_notebook_guid,),
            )

            row = cur.fetchone()

            if row is None:
                raise ValueError(
                    f"No local notebooks found for linked notebook {l_notebook_guid}"
                )

            return Notebook(
                guid=row["guid"],
                name=row["name"],
                stack=row["stack"],
            )

    def get_linked_notebook_usn(self, l_notebook_guid: str) -> int:
        with self.db as con:
            cur = con.execute(
                "select usn from notebooks_linked where guid=?",
                (l_notebook_guid,),
            )

            res = cur.fetchone()

            if res is None:
                return 0

            return int(res[0])

    def set_linked_notebook_usn(self, l_notebook_guid: str, usn: int) -> None:
        with self.db as con:
            con.execute(
                "update notebooks_linked set usn=? where guid=?",
                (usn, l_notebook_guid),
            )

    def expunge_linked_notebooks(self, guids: Iterable[str]) -> None:
        with self.db as con:
            con.executemany(
                "delete from notebooks_linked where guid=?", ((g,) for g in guids)
            )


class NoteStorage(SqliteStorage):  # noqa: WPS214
    def add_notes_for_sync(self, notes: Iterable[Note]) -> None:
        if logger.getEffectiveLevel() == logging.DEBUG:  # pragma: no cover
            for note in notes:
                n_info = log_format_note(note)
                logger.debug(f"Scheduling note for sync {n_info}")

        with self.db as con:
            con.executemany(
                "replace into notes(guid, title, notebook_guid) values (?, ?, ?)",
                ((n.guid, n.title, n.notebookGuid) for n in notes),
            )

    def add_note(self, note: Note) -> None:
        if logger.getEffectiveLevel() == logging.DEBUG:  # pragma: no cover
            n_info = log_format_note(note)
            logger.debug(f"Adding/updating note {n_info}")

        note_deflated = lzma.compress(pickle.dumps(note))

        with self.db as con:
            con.execute(
                "replace into notes(guid, title, notebook_guid, is_active, raw_note)"
                " values (?, ?, ?, ?, ?)",
                (
                    note.guid,
                    note.title,
                    note.notebookGuid,
                    note.active,
                    note_deflated,
                ),
            )

        logger.debug(f"Added note [{note.guid}]")

    def iter_notes(self, notebook_guid: str) -> Iterator[Note]:
        for note_guid in self._get_notes_by_notebook(notebook_guid):
            with self.db as con:
                cur = con.execute(
                    "select raw_note from notes where guid=?", (note_guid,)
                )

                row = cur.fetchone()

                yield pickle.loads(lzma.decompress(row["raw_note"]))

    def iter_notes_trash(self) -> Iterator[Note]:
        with self.db as con:
            cur = con.execute(
                "select raw_note"
                " from notes where is_active=0"
                " order by title COLLATE NOCASE",
            )

            yield from (pickle.loads(lzma.decompress(row["raw_note"])) for row in cur)

    def get_notes_for_sync(self) -> Tuple[NoteForSync, ...]:
        with self.db as con:
            cur = con.execute(
                "select notes.guid, title, notebooks_linked.guid as l_notebook"
                " from notes"
                " left join notebooks_linked"
                " using (notebook_guid)"
                " where raw_note is NULL"
            )

            notes = (
                NoteForSync(
                    guid=row["guid"],
                    title=row["title"],
                    linked_notebook_guid=row["l_notebook"],
                )
                for row in cur.fetchall()
            )

            return tuple(notes)

    def expunge_notes(self, guids: Iterable[str]) -> None:
        with self.db as con:
            con.executemany("delete from notes where guid=?", ((g,) for g in guids))

    def expunge_notes_by_notebook(self, notebook_guid: str) -> None:
        with self.db as con:
            con.execute("delete from notes where notebook_guid=?", (notebook_guid,))

    def get_notes_count(self, is_active: bool = True) -> int:
        with self.db as con:
            cur = con.execute(
                "select COUNT(guid) from notes where is_active=?", (is_active,)
            )

            return int(cur.fetchone()[0])

    def _get_notes_by_notebook(self, notebook_guid: str) -> List[str]:
        """Due to wrong idx_notes index, SQLite creates a temporary table on
            from notes where notebook_guid=? and is_active=1
            order by title COLLATE NOCASE
        which may cause a memory leak. This method sorts notes alphabetically
        to prevent SQLite from creating a sort table."""

        with self.db as con:
            cur = con.execute(
                "select guid, title"
                " from notes where notebook_guid=? and is_active=1",
                (notebook_guid,),
            )

            sorted_notes = sorted(cur, key=lambda x: x["title"])  # type: ignore

            return [r["guid"] for r in sorted_notes]


class ConfigStorage(SqliteStorage):
    def set_config_value(self, name: str, config_value: str) -> None:
        with self.db as con:
            con.execute(
                "replace into config(name, value) values (?, ?)",
                (name, config_value),
            )

    def get_config_value(self, name: str) -> str:
        with self.db as con:
            cur = con.execute("select value from config where name=?", (name,))
            res = cur.fetchone()

            if not res:
                raise KeyError(f"Config ID {name} not found in database!")

            return str(res[0])



import os
from typing import Dict

MAX_FILE_NAME_LEN = 255


class SafePath(object):
    """Ensures path for tuples of directory names that may contain bad symbols

    Use case:
    safe_path.get("test_dir") == base_dir/test_dir
    safe_path.get("test_dir?|") == base_dir/test_dir__
    safe_path.get("test_dir||") == base_dir/test_dir__ (1)
    """

    def __init__(self, base_dir: str) -> None:
        self.main_base_dir = base_dir
        self.safe_paths: Dict[tuple, str] = {}

    def get_file(self, *paths: str) -> str:
        return self.get(*paths, is_dir=False)

    def get(self, *paths: str, is_dir: bool = True) -> str:
        if paths in self.safe_paths:
            return self.safe_paths[paths]  # noqa: WPS529

        if len(paths) > 1:
            parent_dir = paths[:-1]
            base_dir = self.get(*parent_dir)
        else:
            base_dir = _verify_path(self.main_base_dir)

        this_dir = paths[-1]

        safe_path = _get_safe_path(base_dir, this_dir)

        if is_dir:
            os.mkdir(safe_path)
            self.safe_paths[paths] = safe_path

        return safe_path


def _get_safe_path(target_dir: str, new_name: str) -> str:
    safe_name = _replace_bad_characters(new_name)

    safe_name = _get_non_existant_name(safe_name, target_dir)

    return os.path.join(target_dir, safe_name)


def _verify_path(output_path: str) -> str:
    output_path = os.path.normpath(output_path)
    if not os.path.exists(output_path) or not os.path.isdir(output_path):
        os.makedirs(output_path)
    return output_path


def _replace_bad_characters(string: str) -> str:
    bad = r'<>:"/\|?*'

    result_string = string
    for b in bad:
        result_string = result_string.replace(b, "_")

    return result_string


def _get_non_existant_name(safe_name: str, target_dir: str) -> str:
    safe_name = _trim_name(safe_name)
    i = 0
    o_name, o_ext = os.path.splitext(safe_name)
    while os.path.exists(os.path.join(target_dir, safe_name)):
        i += 1
        safe_name = f"{o_name} ({i}){o_ext}"
        if len(safe_name) > MAX_FILE_NAME_LEN:
            max_len = MAX_FILE_NAME_LEN - len(f" ({i}){o_ext}")
            o_name = _trim_name(o_name, max_len)
            safe_name = f"{o_name} ({i}){o_ext}"

    return safe_name


def _trim_name(safe_name: str, max_len: int = MAX_FILE_NAME_LEN) -> str:
    """Trim file name to 255 characters while maintaining extension
    255 characters is max file name length on linux and macOS
    Windows has a path limit of 260 characters which includes
    the entire path (drive letter, path, and file name)
    This does not trim the path length, just the file name

    max_len: if provided, trims to this length otherwise MAX_FILE_NAME_LEN

    Raises: ValueError if the file name is too long and cannot be trimmed
    """
    if len(safe_name) <= max_len:
        return safe_name

    name, ext = os.path.splitext(safe_name)

    drop_chars = len(safe_name) - max_len
    trimmed_name = name[:-drop_chars]

    if not ext:
        return trimmed_name
    if len(name) > drop_chars:
        return f"{trimmed_name}{ext}"

    raise ValueError(f"File name is too long but cannot be safely trimmed: {safe_name}")



# -*- coding: utf-8 -*-

import logging
from datetime import datetime
from typing import Iterable, List, Union

from click import progressbar
from evernote.edam.type.ttypes import Note, Notebook

from evernote_backup.cli_app_util import get_progress_output
from evernote_backup.log_util import log_format_note, log_format_notebook
from evernote_backup.note_exporter_util import SafePath
from evernote_backup.note_formatter import NoteFormatter
from evernote_backup.note_storage import SqliteStorage

logger = logging.getLogger(__name__)

ENEX_HEAD = """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE en-export SYSTEM "http://xml.evernote.com/pub/evernote-export3.dtd">
"""
ENEX_TAIL = "</en-export>"


class NothingToExportError(Exception):
    """Raise when database is empty"""


class NoteExporter(object):
    def __init__(self, storage: SqliteStorage, target_dir: str) -> None:
        self.storage = storage
        self.safe_paths = SafePath(target_dir)

    def export_notebooks(
        self, single_notes: bool, export_trash: bool, no_export_date: bool
    ) -> None:
        count_notes = self.storage.notes.get_notes_count()
        count_trash = self.storage.notes.get_notes_count(is_active=False)

        if logger.getEffectiveLevel() == logging.DEBUG:  # pragma: no cover
            logger.debug(f"Notes to export: {count_notes}")
            logger.debug(f"Trashed notes: {count_trash}")
            if single_notes:
                logger.debug("Export mode: single notes")
            else:
                logger.debug("Export mode: notebooks")

        if count_notes == 0 and count_trash == 0:
            raise NothingToExportError

        if count_notes > 0:
            logger.info("Exporting notes...")

            self._export_active(single_notes, no_export_date)

        if count_trash > 0 and export_trash:
            logger.info("Exporting trash...")

            self._export_trash(single_notes, no_export_date)

    def _export_active(self, single_notes: bool, no_export_date: bool) -> None:
        notebooks = tuple(self.storage.notebooks.iter_notebooks())

        with progressbar(
            notebooks,
            show_pos=True,
            file=get_progress_output(),
        ) as notebooks_bar:
            for nb in notebooks_bar:
                if logger.getEffectiveLevel() == logging.DEBUG:  # pragma: no cover
                    nb_info = log_format_notebook(nb)
                    logger.debug(f"Exporting notebook {nb_info}")

                if self.storage.notebooks.get_notebook_notes_count(nb.guid) == 0:
                    logger.debug("Notebook is empty, skip")
                    continue

                self._export_notes(nb, single_notes, no_export_date)

    def _export_notes(
        self, notebook: Notebook, single_notes: bool, no_export_date: bool
    ) -> None:
        parent_dir = [notebook.stack] if notebook.stack else []

        notes_source = self.storage.notes.iter_notes(notebook.guid)

        if single_notes:
            parent_dir.append(notebook.name)
            self._output_single_notes(parent_dir, notes_source, no_export_date)
        else:
            self._output_notebook(
                parent_dir, notebook.name, notes_source, no_export_date
            )

    def _export_trash(self, single_notes: bool, no_export_date: bool) -> None:
        notes_source = self.storage.notes.iter_notes_trash()

        if single_notes:
            self._output_single_notes(["Trash"], notes_source, no_export_date)
        else:
            self._output_notebook([], "Trash", notes_source, no_export_date)

    def _output_single_notes(
        self,
        parent_dir: List[str],
        notes_source: Iterable[Note],
        no_export_date: bool,
    ) -> None:
        for note in notes_source:
            note_path = self.safe_paths.get_file(*parent_dir, f"{note.title}.enex")

            _write_export_file(note_path, note, no_export_date)

    def _output_notebook(
        self,
        parent_dir: List[str],
        notebook_name: str,
        notes_source: Iterable[Note],
        no_export_date: bool,
    ) -> None:
        notebook_path = self.safe_paths.get_file(*parent_dir, f"{notebook_name}.enex")

        _write_export_file(notebook_path, notes_source, no_export_date)


def _write_export_file(
    file_path: str, note_source: Union[Iterable[Note], Note], no_export_date: bool
) -> None:
    with open(file_path, "w", encoding="utf-8") as f:
        logger.debug(f"Writing file {file_path}")

        f.write(ENEX_HEAD)

        if no_export_date:
            f.write('<en-export application="Evernote" version="10.10.5">')
        else:
            now = datetime.utcnow().strftime("%Y%m%dT%H%M%SZ")
            f.write(
                f'<en-export export-date="{now}"'
                f' application="Evernote" version="10.10.5">'
            )

        note_formatter = NoteFormatter()

        if isinstance(note_source, Note):
            if logger.getEffectiveLevel() == logging.DEBUG:  # pragma: no cover
                n_info = log_format_note(note_source)
                logger.debug(f"Exporting note {n_info}")
            f.write(note_formatter.format_note(note_source))
        else:
            for note in note_source:  # noqa: WPS440
                if logger.getEffectiveLevel() == logging.DEBUG:  # pragma: no cover
                    n_info = log_format_note(note)
                    logger.debug(f"Exporting note {n_info}")
                f.write(note_formatter.format_note(note))

        f.write(ENEX_TAIL)



import logging
import sys
import traceback
from typing import Optional

import click
from click_option_group import MutuallyExclusiveOptionGroup, optgroup
from evernote.edam.error.ttypes import EDAMErrorCode, EDAMSystemException

from evernote_backup import cli_app, config_defaults
from evernote_backup.cli_app_click_util import (
    DIR_ONLY,
    FILE_ONLY,
    NaturalOrderGroup,
    group_options,
)
from evernote_backup.cli_app_util import ProgramTerminatedError
from evernote_backup.log_util import get_time_txt, init_logging, init_logging_format
from evernote_backup.version import __version__

logger = logging.getLogger()


opt_user = click.option(
    "--user",
    "-u",
    help="Account email or username.",
)

opt_password = click.option(
    "--password",
    "-p",
    help="Account password.",
)

opt_oauth = click.option(
    "--oauth",
    is_flag=True,
    help=(
        "OAuth login flow."
        " Use this if you signed up for Evernote with Google or Apple account."
        " (Ignores '--user' and '--password' when used."
        " Doesn't work for Yinxiang.)"
    ),
)

opt_oauth_port = click.option(
    "--oauth-port",
    default=config_defaults.OAUTH_LOCAL_PORT,
    show_default=True,
    help="OAuth local server port. (Advanced option, use with --oauth.)",
)

opt_oauth_host = click.option(
    "--oauth-host",
    default=config_defaults.OAUTH_HOST,
    show_default=True,
    help="OAuth local server host. (Advanced option, use with --oauth.)",
)

opt_token = click.option(
    "--token",
    "-t",
    help=(
        "Manually provide authentication token to use with Evernote API."
        " (Advanced option, ignores '--user', '--password' and '--oauth' when used.)"
    ),
)

opt_network_retry_count = click.option(
    "--network-retry-count",
    default=config_defaults.NETWORK_ERROR_RETRY_COUNT,
    show_default=True,
    type=click.IntRange(1),
    help=("Network error retry count. (Advanced option)"),
)

opt_database = click.option(
    "--database",
    "-d",
    default=config_defaults.DATABASE_NAME,
    show_default=True,
    required=True,
    type=FILE_ONLY,
    help="Database file where notes are stored.",
)


@click.group(cls=NaturalOrderGroup)
@optgroup.group("Verbosity", cls=MutuallyExclusiveOptionGroup)  # type: ignore
@optgroup.option(  # type: ignore
    "--quiet",
    "-q",
    is_flag=True,
    help="Quiet mode, output only critical errors.",
)
@optgroup.option(  # type: ignore
    "--verbose",
    "-v",
    is_flag=True,
    help="Verbose mode, output debug information.",
)
@click.version_option(__version__)
def cli(quiet: bool, verbose: bool) -> None:
    """Evernote backup & export

    \b
    On the first run, to export notes:
    evernote-backup init-db
    evernote-backup sync
    evernote-backup export output_dir/
    """

    init_logging()
    init_logging_format()

    if quiet:
        logger.setLevel(logging.CRITICAL)
    elif verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)


@cli.command()
@opt_database
@group_options(
    opt_user, opt_password, opt_oauth, opt_oauth_port, opt_oauth_host, opt_token
)
@click.option(
    "--force",
    is_flag=True,
    help="Overwrite existing database file.",
)
@click.option(
    "--backend",
    default=config_defaults.BACKEND,
    show_default=True,
    type=click.Choice(["evernote", "evernote:sandbox", "china", "china:sandbox"]),
    help="API backend to connect to. If you are using Yinxiang, select 'china'.",
)
@opt_network_retry_count
def init_db(
    database: str,
    user: Optional[str],
    password: Optional[str],
    oauth: bool,
    oauth_port: int,
    oauth_host: str,
    token: Optional[str],
    force: bool,
    backend: str,
    network_retry_count: int,
) -> None:
    """Initialize storage & log in to Evernote."""

    cli_app.init_db(
        database=database,
        auth_user=user,
        auth_password=password,
        auth_is_oauth=oauth,
        auth_oauth_port=oauth_port,
        auth_oauth_host=oauth_host,
        auth_token=token,
        force=force,
        backend=backend,
        network_retry_count=network_retry_count,
    )


@cli.command()
@opt_database
@click.option(
    "--max-chunk-results",
    default=config_defaults.SYNC_CHUNK_MAX_RESULTS,
    type=click.IntRange(1, config_defaults.SYNC_CHUNK_MAX_RESULTS_SERVER_LIMIT),
    show_default=True,
    help="Max entries per sync chunk. (Advanced option)",
)
@click.option(
    "--max-download-workers",
    default=config_defaults.SYNC_MAX_DOWNLOAD_WORKERS,
    show_default=True,
    type=click.IntRange(1, config_defaults.SYNC_MAX_DOWNLOAD_WORKERS_SANE_LIMIT),
    help=(
        "Max number of parallel downloads. Don't set too high to avoid rate limits."
        " (Advanced option)"
    ),
)
@click.option(
    "--download-cache-memory-limit",
    default=config_defaults.SYNC_DOWNLOAD_CACHE_MEMORY_LIMIT,
    show_default=True,
    type=click.IntRange(1),
    help=(
        "Cache size in MB for notes stored in memory before writing to disk."
        " (Advanced option)"
    ),
)
@opt_network_retry_count
def sync(
    database: str,
    max_chunk_results: int,
    max_download_workers: int,
    download_cache_memory_limit: int,
    network_retry_count: int,
) -> None:
    """Sync local database with Evernote, downloading all notes."""

    cli_app.sync(
        database=database,
        max_chunk_results=max_chunk_results,
        max_download_workers=max_download_workers,
        download_cache_memory_limit=download_cache_memory_limit,
        network_retry_count=network_retry_count,
    )


@cli.command()
@opt_database
@click.option(
    "--single-notes",
    is_flag=True,
    help="Export single notes instead of notebooks.",
)
@click.option(
    "--include-trash",
    is_flag=True,
    help="Include notes from trash into export.",
)
@click.option(
    "--no-export-date",
    is_flag=True,
    help=(
        "Don't timestamp exported enex files."
        " (e.g. to prevent backup chunking with zero changes)"
    ),
)
@click.argument(
    "output_path",
    required=True,
    type=DIR_ONLY,
)
def export(
    database: str,
    single_notes: bool,
    include_trash: bool,
    no_export_date: bool,
    output_path: str,
) -> None:
    """Export all notes from local database into ENEX files."""

    cli_app.export(
        database=database,
        single_notes=single_notes,
        include_trash=include_trash,
        no_export_date=no_export_date,
        output_path=output_path,
    )


click.password_option()


@cli.command()
@opt_database
@group_options(
    opt_user, opt_password, opt_oauth, opt_oauth_port, opt_oauth_host, opt_token
)
@opt_network_retry_count
def reauth(
    database: str,
    user: Optional[str],
    password: Optional[str],
    oauth: bool,
    oauth_port: int,
    oauth_host: str,
    token: Optional[str],
    network_retry_count: int,
) -> None:
    """Refresh login to Evernote, run when token expires."""

    cli_app.reauth(
        database=database,
        auth_user=user,
        auth_password=password,
        auth_is_oauth=oauth,
        auth_oauth_port=oauth_port,
        auth_oauth_host=oauth_host,
        auth_token=token,
        network_retry_count=network_retry_count,
    )


def main() -> None:
    try:
        cli()
    except ProgramTerminatedError as e:
        logger.critical(e)
        sys.exit(1)
    except EDAMSystemException as e:
        if e.errorCode != EDAMErrorCode.RATE_LIMIT_REACHED:
            logger.critical(traceback.format_exc())
            sys.exit(1)

        time_left = get_time_txt(e.rateLimitDuration)

        logger.critical(f"Rate limit reached. Restart program in {time_left}.")
        sys.exit(1)
    except Exception:
        logger.critical(traceback.format_exc())
        sys.exit(1)
